﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;

using Microsoft.Kinect;

namespace FrameSupplier
{

    class FrameSupplier
    {
        private KinectSensor sensor;

        private MultiSourceFrameReader multiFrameReader;

        private BodyFrame bodyFrame;
        private ColorFrame colorFrame;
        private DepthFrame depthFrame;
        private InfraredFrame infraredFrame;

        private byte[] bodymessage;
        private byte[] colormessage;
        private byte[] depthmessage;
        private byte[] infraredmessage;

        private Body[] bodies;

        public FrameSupplier()
        {
            sensor = KinectSensor.GetDefault();

            if (sensor != null)
            {
                Console.WriteLine("Setting up the Kinect sensor");
                sensor.Open();
            }
            else
            {
                Console.WriteLine("Setting up the Kinect sensor failed");
                return;
            }

            Console.WriteLine("Kinect sensor set up successfully");

            Console.WriteLine("Setting up the frame reader");
            multiFrameReader = sensor.OpenMultiSourceFrameReader(FrameSourceTypes.Color | 
                                                                 FrameSourceTypes.Depth | 
                                                                 FrameSourceTypes.Infrared |
                                                                 FrameSourceTypes.Body);

            multiFrameReader.MultiSourceFrameArrived += MultiFrameReader_FrameArrived;

            Console.WriteLine("Frame reader set up successfully");

            bodyFrame = null;
            colorFrame = null;
            depthFrame = null;
            infraredFrame = null;

            bodies = new Body[6];

            bodymessage = new byte[6 * 4];
            colormessage = new byte[1920 * 1080 * ((PixelFormats.Bgr32.BitsPerPixel + 7) / 8)];
            depthmessage = new byte[512 * 424 * ((PixelFormats.Bgr32.BitsPerPixel + 7) / 8)];
            infraredmessage = new byte[512 * 424 * ((PixelFormats.Bgr32.BitsPerPixel + 7) / 8)];

            Console.WriteLine("\n--------------------------------------" +
                              "\n        Sensor setup completed" +
                              "\n--------------------------------------\n");

        }
        
        public bool SensorOn()
        {
            if (sensor == null)
            {
                return false;
            }
            return true;
        }

        public byte[] GetByteArray(string requiredFrame)
        {
            if (requiredFrame == "body")
            {
                
                var frame = bodyFrame;
                if (frame != null)
                {
                    Console.WriteLine("Body frame acquired");
                    return bodymessage;
                }
                else
                {
                    Console.WriteLine("Could not acquire bodyframe");
                }
            }
            else if (requiredFrame == "color")
            {
                var frame = colorFrame;
                if (frame != null)
                {
                    Console.WriteLine("Color frame acquired");
                    return colormessage;
                }
                else
                {
                    Console.WriteLine("Could not acquire colorframe");
                }
            }
            else if (requiredFrame == "depth")
            {
                var frame = depthFrame;
                if (frame != null)
                {
                    Console.WriteLine("Depth frame acquired");
                    return depthmessage;
                }
                else
                {
                    Console.WriteLine("Could not acquire depthframe");
                }
            }
            else if (requiredFrame == "infrared")
            {
                var frame = infraredFrame;
                if (frame != null)
                {
                    Console.WriteLine("Infrared frame acquired");
                    return infraredmessage;
                }
                else
                {
                    Console.WriteLine("Could not acquire infraredframe");
                }
            }
            return null;
        }

        private byte[] ToByteArray(BodyFrame frame)
        {
            byte[] heads = new byte[6 * 4];
            if (bodies != null)
            {
                
                int i = 0;

                if (frame != null)
                {
                    foreach (Body body in bodies)
                    {
                        heads[i++] = (byte)body.TrackingId;
                        heads[i++] = (byte)body.Joints[JointType.Head].Position.X;
                        heads[i++] = (byte)body.Joints[JointType.Head].Position.Y;
                        heads[i++] = (byte)body.Joints[JointType.Head].Position.Z;
                    }
                }

            }

            return heads;
        }

        private byte[] ToByteArray(ColorFrame frame)
        {
            if (frame != null)
            {
                int width = frame.FrameDescription.Width;
                int height = frame.FrameDescription.Height;

                byte[] pixels = new byte[width * height * ((PixelFormats.Bgr32.BitsPerPixel + 7) / 8)];

                if (frame.RawColorImageFormat == ColorImageFormat.Bgra)
                {
                    frame.CopyRawFrameDataToArray(pixels);
                }
                else
                {
                    frame.CopyConvertedFrameDataToArray(pixels, ColorImageFormat.Bgra);
                }

                return pixels;
            }
            return new byte[6];
        }

        private byte[] ToByteArray(DepthFrame frame)
        {
            int width = frame.FrameDescription.Width;
            int height = frame.FrameDescription.Height;

            ushort minDepth = frame.DepthMinReliableDistance;
            ushort maxDepth = frame.DepthMaxReliableDistance;

            ushort[] depthData = new ushort[width * height];
            byte[] pixelData = new byte[width * height * ((PixelFormats.Bgr32.BitsPerPixel + 7) / 8)];

            frame.CopyFrameDataToArray(depthData);

            int colorIndex = 0;
            for (int depthIndex = 0; depthIndex < depthData.Length; depthIndex++)
            {
                ushort depth = depthData[depthIndex];
                byte intensity = (byte)(depth >= minDepth && depth <= maxDepth ? depth : 0);

                pixelData[colorIndex++] = intensity;
                pixelData[colorIndex++] = intensity;
                pixelData[colorIndex++] = intensity;

                colorIndex++;
            }

            return pixelData;
        }

        private byte[] ToByteArray(InfraredFrame frame)
        {
            int width = frame.FrameDescription.Width;
            int height = frame.FrameDescription.Height;

            ushort[] infraredData = new ushort[width * height];
            byte[] pixelData = new byte[width * height * ((PixelFormats.Bgr32.BitsPerPixel + 7) / 8)];

            frame.CopyFrameDataToArray(infraredData);

            int colorIndex = 0;
            for (int infraredIndex = 0; infraredIndex < infraredData.Length; infraredIndex++)
            {
                ushort ir = infraredData[infraredIndex];
                byte intensity = (byte)(ir >> 9);

                pixelData[colorIndex++] = intensity; //Blue
                pixelData[colorIndex++] = intensity; //Green
                pixelData[colorIndex++] = intensity; //Red

                colorIndex++;
            }

            return pixelData;
        }

        void MultiFrameReader_FrameArrived(object sender, MultiSourceFrameArrivedEventArgs args)
        {
            var reference = args.FrameReference.AcquireFrame();

            using (var frame = reference.ColorFrameReference.AcquireFrame())
            {
                if (frame != null)
                {
                    colorFrame = frame;
                    colormessage = ToByteArray(frame);
                }
            }
            using (var frame = reference.DepthFrameReference.AcquireFrame())
            {
                if (frame != null)
                {
                    depthFrame = frame;
                    depthmessage = ToByteArray(frame);
                }
            }
            using (var frame = reference.InfraredFrameReference.AcquireFrame())
            {
                if (frame != null)
                {
                    infraredFrame = frame;
                    infraredmessage = ToByteArray(frame);
                }
            }
            using (var frame = reference.BodyFrameReference.AcquireFrame())
            {
                if (frame != null)
                {
                    bodyFrame = frame;
                    bodyFrame.GetAndRefreshBodyData(bodies);
                    bodymessage = ToByteArray(frame);
                }
            }
        }

    }
}
