﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;

using Microsoft.Kinect;

namespace FrameSupplier
{

    class FrameSupplier
    {
        private KinectSensor sensor;

        private BodyFrameReader bodyReader;
        private ColorFrameReader colorReader;
        private DepthFrameReader depthReader;
        private InfraredFrameReader infraredReader;

        private BodyFrame bodyFrame;
        private ColorFrame colorFrame;
        private DepthFrame depthFrame;
        private InfraredFrame infraredFrame;

        private Body[] bodies;

        public FrameSupplier()
        {
            sensor = KinectSensor.GetDefault();

            if (sensor != null)
            {
                Console.WriteLine("setting up sensor");
                sensor.Open();
            }
            else
            {
                return;
            }

            bodyReader = sensor.BodyFrameSource.OpenReader();
            colorReader = sensor.ColorFrameSource.OpenReader();
            depthReader = sensor.DepthFrameSource.OpenReader();
            infraredReader = sensor.InfraredFrameSource.OpenReader();

            bodyReader.FrameArrived += BodyReader_FrameArrived;
            colorReader.FrameArrived += ColorReader_FrameArrived;
            depthReader.FrameArrived += DepthReader_FrameArrived;
            infraredReader.FrameArrived += InfraredReader_FrameArrived;

            bodies = new Body[6];
        }

        ~FrameSupplier()
        {
            if (bodyReader != null)
            {
                bodyReader.Dispose();
                bodyReader = null;
            }
            if (colorReader != null)
            {
                colorReader.Dispose();
                colorReader = null;
            }
            if (depthReader != null)
            {
                depthReader.Dispose();
                depthReader = null;
            }
            if (infraredReader != null)
            {
                infraredReader.Dispose();
                infraredReader = null;
            }
        }

        public byte[] HandleRequest(byte[] request)
        {
            return GetByteArray(Convert.ToString(request));
        }

        private byte[] GetByteArray(string requiredFrame)
        {
            if (requiredFrame == "body")
            {
                var frame = bodyFrame;
                if (frame != null)
                {
                    Console.WriteLine("Body frame acquired");
                    return ToByteArray(frame);
                }
                else
                {
                    Console.WriteLine("Could not acquire bodyframe");
                }
            }
            else if (requiredFrame == "color")
            {
                var frame = colorFrame;
                if (frame != null)
                {
                    Console.WriteLine("Color frame acquired");
                    return ToByteArray(frame);
                }
                else
                {
                    Console.WriteLine("Could not acquire colorframe");
                }
            }
            else if (requiredFrame == "depth")
            {
                var frame = depthFrame;
                if (frame != null)
                {
                    Console.WriteLine("Depth frame acquired");
                    return ToByteArray(frame);
                }
                else
                {
                    Console.WriteLine("Could not acquire depthframe");
                }
            }
            else if (requiredFrame == "infrared")
            {
                var frame = infraredFrame;
                if (frame != null)
                {
                    Console.WriteLine("Infrared frame acquired");
                    return ToByteArray(frame);
                }
                else
                {
                    Console.WriteLine("Could not acquire infraredframe");
                }
            }
            return null;
        }

        private byte[] ToByteArray(BodyFrame frame)
        {
            byte[] heads = null;
            int i = 0;

            if (frame != null)
            {
                try
                {
                    frame.GetAndRefreshBodyData(bodies);
                    heads = new byte[bodies.Length * 4];
                    foreach(Body body in bodies){
                        heads[i++] = (byte)body.TrackingId;
                        heads[i++] = (byte)body.Joints[JointType.Head].Position.X;
                        heads[i++] = (byte)body.Joints[JointType.Head].Position.Y;
                        heads[i++] = (byte)body.Joints[JointType.Head].Position.Z;
                    }
                }
                catch (System.ArgumentOutOfRangeException e)
                {

                }
            }
            return heads;
        }

        private byte[] ToByteArray(ColorFrame frame)
        {
            int width = frame.FrameDescription.Width;
            int height = frame.FrameDescription.Height;

            byte[] pixels = new byte[width * height * ((PixelFormats.Bgr32.BitsPerPixel + 7) / 8)];

            if (frame.RawColorImageFormat == ColorImageFormat.Bgra)
            {
                frame.CopyRawFrameDataToArray(pixels);
            }
            else
            {
                frame.CopyConvertedFrameDataToArray(pixels, ColorImageFormat.Bgra);
            }

            return pixels;
        }

        private byte[] ToByteArray(DepthFrame frame)
        {
            int width = frame.FrameDescription.Width;
            int height = frame.FrameDescription.Height;

            ushort minDepth = frame.DepthMinReliableDistance;
            ushort maxDepth = frame.DepthMaxReliableDistance;

            ushort[] depthData = new ushort[width * height];
            byte[] pixelData = new byte[width * height * ((PixelFormats.Bgr32.BitsPerPixel + 7) / 8)];

            frame.CopyFrameDataToArray(depthData);

            int colorIndex = 0;
            for (int depthIndex = 0; depthIndex < depthData.Length; depthIndex++)
            {
                ushort depth = depthData[depthIndex];
                byte intensity = (byte)(depth >= minDepth && depth <= maxDepth ? depth : 0);

                pixelData[colorIndex++] = intensity;
                pixelData[colorIndex++] = intensity;
                pixelData[colorIndex++] = intensity;

                colorIndex++;
            }

            return pixelData;
        }

        private byte[] ToByteArray(InfraredFrame frame)
        {
            int width = frame.FrameDescription.Width;
            int height = frame.FrameDescription.Height;

            ushort[] infraredData = new ushort[width * height];
            byte[] pixelData = new byte[width * height * ((PixelFormats.Bgr32.BitsPerPixel + 7) / 8)];

            frame.CopyFrameDataToArray(infraredData);

            int colorIndex = 0;
            for (int infraredIndex = 0; infraredIndex < infraredData.Length; infraredIndex++)
            {
                ushort ir = infraredData[infraredIndex];
                byte intensity = (byte)(ir >> 9);

                pixelData[colorIndex++] = intensity; //Blue
                pixelData[colorIndex++] = intensity; //Green
                pixelData[colorIndex++] = intensity; //Red

                colorIndex++;
            }

            return pixelData;
        }

        static void Main()
        {
            byte[] request = {Convert.ToByte('b'), Convert.ToByte('o'), Convert.ToByte('d'), Convert.ToByte('y')};

            FrameSupplier fs = new FrameSupplier();
            /*Sleep a bit to let frames be aqcuired*/
            Thread.Sleep(1500);
            fs.HandleRequest(request);
        }

        void BodyReader_FrameArrived(object sender, BodyFrameArrivedEventArgs args)
        {
            bodyFrame = args.FrameReference.AcquireFrame();
        }

        void ColorReader_FrameArrived(object sender, ColorFrameArrivedEventArgs args)
        {
            colorFrame = args.FrameReference.AcquireFrame();
        }

        void DepthReader_FrameArrived(object sender, DepthFrameArrivedEventArgs args)
        {
            depthFrame = args.FrameReference.AcquireFrame();
        }

        void InfraredReader_FrameArrived(object sender, InfraredFrameArrivedEventArgs args)
        {
            infraredFrame = args.FrameReference.AcquireFrame();
        }

    }
}
